/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import java.util.*;
import java.util.function.*;

/**
 * Definiert mathematische Funktionen auf Zugehörigkeiten.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public final class Funktionen {

    private Funktionen() {
        // utility class
    }

    public static Zugehoerigkeit minimum(Zugehoerigkeit a, Zugehoerigkeit b) {
        Objects.requireNonNull(a, "a");
        Objects.requireNonNull(b, "b");
        if (a.isNull() || b.isNull()) {
            return Zugehoerigkeit.NULL;
        }

        return Zugehoerigkeit.von(Math.min(a.getWert(), b.getWert()));
    }

    public static Zugehoerigkeit maximum(Zugehoerigkeit a, Zugehoerigkeit b) {
        Objects.requireNonNull(a, "a");
        Objects.requireNonNull(b, "b");
        if (a.isNull() || b.isNull()) {
            return Zugehoerigkeit.NULL;
        }

        return Zugehoerigkeit.von(Math.max(a.getWert(), b.getWert()));
    }

    public static Zugehoerigkeit komplement(Zugehoerigkeit a) {
        Objects.requireNonNull(a, "a");
        if (a.isNull()) {
            return Zugehoerigkeit.NULL;
        }

        return Zugehoerigkeit.von(1.0 - a.getWert());
    }

    public static Zugehoerigkeit implikation(Zugehoerigkeit a, Zugehoerigkeit b) {
        return maximum(komplement(a), b);
    }

    public static Zugehoerigkeit algebraischesProdukt(Zugehoerigkeit a, Zugehoerigkeit b) {
        Objects.requireNonNull(a, "a");
        Objects.requireNonNull(b, "b");
        if (a.isNull() || b.isNull()) {
            return Zugehoerigkeit.NULL;
        }

        return Zugehoerigkeit.von(a.getWert() * b.getWert());
    }

    public static Zugehoerigkeit algebraischesSumme(Zugehoerigkeit a, Zugehoerigkeit b) {
        Objects.requireNonNull(a, "a");
        Objects.requireNonNull(b, "b");
        if (a.isNull() || b.isNull()) {
            return Zugehoerigkeit.NULL;
        }

        return Zugehoerigkeit.von(a.getWert() + b.getWert() - (a.getWert() * b.getWert()));
    }

    public static Zugehoerigkeit gammaOperator(Zugehoerigkeit a, Zugehoerigkeit b, Zugehoerigkeit gamma) {
        Objects.requireNonNull(a, "a");
        Objects.requireNonNull(b, "b");
        Objects.requireNonNull(gamma, "gamma");
        if (a.isNull() || b.isNull() || gamma.isNull()) {
            return Zugehoerigkeit.NULL;
        }

        double a1 = a.getWert();
        double b1 = b.getWert();
        double gamma1 = gamma.getWert();
        double z1 = Math.pow(a1 * b1, 1 - gamma1) * Math.pow((a1 + b1) - (a1 * b1), gamma1);
        return Zugehoerigkeit.von(z1);
    }

    public static BiFunction<Zugehoerigkeit, Zugehoerigkeit, Zugehoerigkeit> gammaOperator(Zugehoerigkeit gamma) {
        Objects.requireNonNull(gamma, "gamma");
        if (gamma.isNull()) {
            return (a, b) -> Zugehoerigkeit.NULL;
        }

        return (a, b) -> gammaOperator(a, b, gamma);
    }

}
