/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import java.util.*;

/**
 * Ein Fuzzy-Set ist eine unscharfe Menge definiert durch eine Trapezfunktion.
 * <p>
 * Die Trapezecken t1, t2, t3 und t4 sind monoton aufsteigend sortiert, wobei t1
 * und t4 stets die Zugehörigkeit 0.0 sowie t2 und t3 stets die Zugehörigkeit
 * 1.0 besitzen. Damit lassen sich die Funktionen Trapez, Rechteck, Dreieck,
 * Strich, aufsteigende Rampe und abfallende Rampe abbilden.
 * <p>
 * Es gibt einen Spezialfall. Bei einem Fuzzy-Set mit gleitenden Übergängen,
 * werden nur t2 und t3 definiert. Eine linguistische Variable muss t1 und t4 so
 * bestimmen, dass zwischen den Fuzzy-Sets gleitende Übergänge entstehen.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public final class FuzzySet {

    private final String name;
    private final double t1;
    private final double t2;
    private final double t3;
    private final double t4;

    private FuzzySet(String name, double t1, double t2, double t3, double t4) {
        this.name = Objects.requireNonNull(name, "name");
        this.t1 = t1;
        this.t2 = t2;
        this.t3 = t3;
        this.t4 = t4;
    }

    public static FuzzySet trapez(String name, double t1, double t2, double t3, double t4) {
        if (t1 > t2 || t2 > t3 || t3 > t4) {
            throw new IllegalArgumentException("Die vier Trapezecken müssen monoton aufsteigend sortiert angegeben werden: " + t1 + ", " + t2 + ", " + t3 + ", " + t4);
        }
        return new FuzzySet(name, t1, t2, t3, t4);
    }

    public static FuzzySet dreieck(String name, double t1, double t2, double t4) {
        return trapez(name, t1, t2, t2, t4);
    }

    public static FuzzySet strich(String name, double t1) {
        return trapez(name, t1, t1, t1, t1);
    }

    public static FuzzySet aufsteigendeRampe(String name, double t1, double t2, double t4) {
        return trapez(name, t1, t2, t4, t4);
    }

    public static FuzzySet abfallendeRampe(String name, double t1, double t3, double t4) {
        return trapez(name, t1, t1, t3, t4);
    }

    public static FuzzySet rechteck(String name, double t1, double t4) {
        return trapez(name, t1, t1, t4, t4);
    }

    public static FuzzySet trapezMitGleitendenUebergaengen(String name, double t2, double t3) {
        if (t2 > t3) {
            throw new IllegalArgumentException("Gleitende Übergänge: Die beiden Trapezecken müssen monoton aufsteigend sortiert angegeben werden: " + t2 + ", " + t3);
        }
        return new FuzzySet(name, Double.NaN, t2, t3, Double.NaN);
    }

    public String getName() {
        return name;
    }

    public double getT1() {
        return t1;
    }

    public double getT2() {
        return t2;
    }

    public double getT3() {
        return t3;
    }

    public double getT4() {
        return t4;
    }

    /**
     * Berechnet mit Hilfe des Strahlensatzes die Zugehörigkeit eines Terms.
     *
     * <p><img src="doc-files/strahlensatz.png" alt="Strahlensatz"></p>
     */
    double bestimmeZugehoerigkeit(double wert) {
        if (t1 < wert && wert < t2) {
            double laengeZA = wert - t1;
            double laengeZC = t2 - t1;
            double laengeCD = 1.0;
            return (laengeZA / laengeZC) * laengeCD;
        } else if (t2 <= wert && wert <= t3) {
            return 1.0;
        } else if (t3 < wert && wert < t4) {
            double laengeZA = t4 - wert;
            double laengeZC = t4 - t3;
            double laengeCD = 1.0;
            return (laengeZA / laengeZC) * laengeCD;
        } else {
            return 0.0;
        }
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        FuzzySet fuzzySet = (FuzzySet) o;
        return Double.compare(fuzzySet.t1, t1) == 0 &&
                Double.compare(fuzzySet.t2, t2) == 0 &&
                Double.compare(fuzzySet.t3, t3) == 0 &&
                Double.compare(fuzzySet.t4, t4) == 0 &&
                Objects.equals(name, fuzzySet.name);
    }

    @Override
    public int hashCode() {
        return Objects.hash(name, t1, t2, t3, t4);
    }

    @Override
    public String toString() {
        return name + ": " + "[" + t1 + ", " + t2 + ", " + t3 + ", " + t4 + "]";
    }

}
