/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import java.util.*;
import java.util.stream.*;

/**
 * Eine linguistische Variable beschreibt die Fuzzy-Sets einer Messgröße.
 * <p>
 * Eine linguistische Variable ist nicht definiert ({@link #isNull()}), wenn sie
 * keine Fuzzy-Sets besitzt.
 * <p>
 * Durch Fuzzyfizierung wird für einen scharfen Messwert eine Fuzzy-Variable
 * bestimmt. Durch Defuzzyfizierung wird für eine Fuzzy-Variable ein scharfer
 * Messwert bestimmt.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 * @see FuzzySet
 * @see FuzzyVariable
 */
public final class LinguistischeVariable {

    private final String name;
    private final Wertebereich wertebereich;
    private final List<FuzzySet> fuzzySets;

    public LinguistischeVariable(String name) {
        this(name, Wertebereich.KEINE_EINSCHRAENKUNG);
    }

    public LinguistischeVariable(String name, Wertebereich wertebereich, FuzzySet... fuzzySet) {
        this(name, wertebereich, Arrays.asList(fuzzySet));
    }

    public LinguistischeVariable(String name, Wertebereich wertebereich, List<FuzzySet> fuzzySets) {
        this.name = Objects.requireNonNull(name, "name");
        this.wertebereich = Objects.requireNonNull(wertebereich, "wertebereich");

        Objects.requireNonNull(fuzzySets, "fuzzySets");
        this.fuzzySets = Collections.unmodifiableList(new ArrayList<>(fuzzySets));
    }

    public String getName() {
        return name;
    }

    public Wertebereich getWertebereich() {
        return wertebereich;
    }

    public boolean isNull() {
        return fuzzySets.isEmpty();
    }

    public boolean nonNull() {
        return !isNull();
    }

    public List<FuzzySet> getFuzzySets() {
        return fuzzySets;
    }

    public List<String> getTerme() {
        return fuzzySets.stream()
                .map(FuzzySet::getName)
                .collect(Collectors.toList());
    }

    public FuzzySet gibFuzzySet(final String name) {
        Objects.requireNonNull(name, "name");
        return fuzzySets.stream()
                .filter(t -> name.equals(t.getName()))
                .findFirst()
                .orElseThrow(() -> new IllegalArgumentException("Die linguistische Variable \"" + this.name + "\" hat kein Fuzzy-Set \"" + name + "\"."));
    }

    public FuzzyVariable erzeugeFuzzyVariable() {
        List<Term> terme = fuzzySets.stream()
                .map(FuzzySet::getName)
                .distinct()
                .map(name -> new Term(name, Zugehoerigkeit.NULL))
                .collect(Collectors.toList());
        return new FuzzyVariable(name, terme);
    }

    public FuzzyVariable fuzzyfiziere(double wert) {
        return new Fuzzyfizierung(mitGleitendenUebergaenge()).fuzzyfiziere(wert);
    }

    public double defuzzyfiziere(FuzzyVariable fuzzyWert) {
        return new Defuzzyfizierung(mitGleitendenUebergaenge()).defuzzyfiziere(fuzzyWert);
    }

    LinguistischeVariable mitGleitendenUebergaenge() {
        List<FuzzySet> fuzzySetsMitUebergaengen = this.fuzzySets.stream()
                .sorted(Comparator.comparingDouble(FuzzySet::getT2))
                .collect(Collectors.toList());
        Wertebereich wertebereich = getWertebereich();
        for (int i = 0; i < fuzzySetsMitUebergaengen.size(); i++) {
            FuzzySet vorgaenger = i > 0 ? fuzzySetsMitUebergaengen.get(i - 1) : null;
            FuzzySet fuzzySet = fuzzySetsMitUebergaengen.get(i);
            FuzzySet nachfolger = i < fuzzySetsMitUebergaengen.size() - 1 ? fuzzySetsMitUebergaengen.get(i + 1) : null;

            double t1;
            if (Double.isNaN(fuzzySet.getT1())) {
                if (vorgaenger == null) {
                    t1 = wertebereich.getMinimum();
                } else {
                    t1 = vorgaenger.getT3();
                }
            } else {
                t1 = fuzzySet.getT1();
            }
            double t2 = fuzzySet.getT2();
            double t3 = fuzzySet.getT3();
            double t4;
            if (Double.isNaN(fuzzySet.getT4())) {
                if (nachfolger == null) {
                    t4 = wertebereich.getMaximum();
                } else {
                    t4 = nachfolger.getT2();
                }
            } else {
                t4 = fuzzySet.getT4();
            }
            fuzzySetsMitUebergaengen.set(i, FuzzySet.trapez(fuzzySet.getName(), t1, t2, t3, t4));
        }
        return new LinguistischeVariable(name, wertebereich, fuzzySetsMitUebergaengen);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        LinguistischeVariable that = (LinguistischeVariable) o;
        return Objects.equals(name, that.name) &&
                Objects.equals(wertebereich, that.wertebereich) &&
                Objects.equals(fuzzySets, that.fuzzySets);
    }

    @Override
    public int hashCode() {
        return Objects.hash(name, wertebereich, fuzzySets);
    }

    @Override
    public String toString() {
        return name + " (" + wertebereich + "): " + fuzzySets;
    }

    private static class Fuzzyfizierung {

        private final LinguistischeVariable linguistischeVariable;

        Fuzzyfizierung(LinguistischeVariable linguistischeVariable) {
            this.linguistischeVariable = linguistischeVariable;
        }

        FuzzyVariable fuzzyfiziere(double wert) {
            List<FuzzySet> fuzzySets = linguistischeVariable.getFuzzySets();
            List<Term> terme = fuzzyfiziere(fuzzySets, wert);
            terme = fasseGleicheTermeZusammen(terme);
            return new FuzzyVariable(linguistischeVariable.getName(), terme);
        }

        private List<Term> fuzzyfiziere(List<FuzzySet> fuzzySets, double wert) {
            return fuzzySets.stream()
                    .map(fs -> bestimmeTerm(fs, wert))
                    .collect(Collectors.toList());
        }

        private List<Term> fasseGleicheTermeZusammen(List<Term> terme) {
            List<Term> zusammengefasst = new ArrayList<>();
            for (Term t : terme) {
                boolean neuerTerm = true;
                for (int i = 0; i < zusammengefasst.size(); i++) {
                    if (t.getName().equals(zusammengefasst.get(i).getName())) {
                        double z1 = t.getZugehoerigkeit().getWert();
                        double z2 = zusammengefasst.get(i).getZugehoerigkeit().getWert();
                        double z = Math.max(z1, z2);
                        zusammengefasst.set(i, new Term(t.getName(), Zugehoerigkeit.von(z)));
                        neuerTerm = false;
                        break;
                    }
                }
                if (neuerTerm) {
                    zusammengefasst.add(t);
                }
            }
            return zusammengefasst;
        }

        private Term bestimmeTerm(FuzzySet fuzzySet, double wert) {
            double z = fuzzySet.bestimmeZugehoerigkeit(wert);
            return new Term(fuzzySet.getName(), Zugehoerigkeit.von(z));
        }

    }

    private static class Defuzzyfizierung {

        private final LinguistischeVariable linguistischeVariable;

        Defuzzyfizierung(LinguistischeVariable linguistischeVariable) {
            this.linguistischeVariable = linguistischeVariable;
        }

        double defuzzyfiziere(FuzzyVariable fuzzyWert) {
            List<FuzzySet> fuzzySets = bestimmeRelevanteFuzzySets(fuzzyWert);
            if (fuzzySets.isEmpty()) {
                return Double.NaN;
            }
            Wertebereich intervall = bestimmeIntervall(fuzzySets);
            return bestimmeFlaechenschwerpunkt(fuzzyWert, fuzzySets, intervall);
        }

        private List<FuzzySet> bestimmeRelevanteFuzzySets(FuzzyVariable fuzzyWert) {
            List<String> gemeinsameTerme = bestimmeGemeinsameTerme(fuzzyWert);
            return gemeinsameTerme.stream()
                    .map(linguistischeVariable::gibFuzzySet)
                    .sorted(Comparator.comparingDouble(FuzzySet::getT1))
                    .collect(Collectors.toList());
        }

        private Wertebereich bestimmeIntervall(List<FuzzySet> fuzzySets) {
            double min = fuzzySets.stream()
                    .mapToDouble(FuzzySet::getT1)
                    .min()
                    .getAsDouble();
            double max = fuzzySets.stream()
                    .mapToDouble(FuzzySet::getT4)
                    .max()
                    .getAsDouble();
            return new Wertebereich(min, max);
        }

        private double bestimmeFlaechenschwerpunkt(FuzzyVariable fuzzyWert, List<FuzzySet> fuzzySets, Wertebereich intervall) {
            double min = intervall.getMinimum();
            double max = intervall.getMaximum();
            double minSchrittweite = 0.5;
            int maxAnzahlStuetzpunkte = 100;
            int anzahlStuetzpunkte = (int) Math.min((max - min) / minSchrittweite, maxAnzahlStuetzpunkte);

            double a = 0;
            double l = 0;
            for (int i = 0; i <= anzahlStuetzpunkte; i++) {
                double stuetzpunkt = min + i * (max - min) / anzahlStuetzpunkte;
                double zugehoerigkeit = Double.NEGATIVE_INFINITY;
                for (FuzzySet fs : fuzzySets) {
                    double zugehoerigkeitFuzzySet = fs.bestimmeZugehoerigkeit(stuetzpunkt);
                    double zugehoerigkeitTerm = fuzzyWert.gibTerm(fs.getName()).getZugehoerigkeit().getWert();
                    double z = Math.min(zugehoerigkeitFuzzySet, zugehoerigkeitTerm);
                    if (z > zugehoerigkeit) {
                        zugehoerigkeit = z;
                    }
                }
                a += stuetzpunkt * zugehoerigkeit;
                l += zugehoerigkeit;
            }
            return a / l;
        }

        private List<String> bestimmeGemeinsameTerme(FuzzyVariable fuzzyVariable) {
            List<String> nonNullTerme = fuzzyVariable.getTerme().stream()
                    .filter(t -> t.getZugehoerigkeit().nonNull())
                    .filter(t -> t.getZugehoerigkeit().getWert() > 0)
                    .map(Term::getName)
                    .collect(Collectors.toList());
            nonNullTerme.retainAll(linguistischeVariable.getTerme());
            return nonNullTerme;
        }

    }


}
