/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import java.util.*;
import java.util.function.*;
import java.util.stream.*;

/**
 * Eine Regelbasis ist eine Menge von Regeln.
 * <p>
 * Die Regelbasis wertet alle enthaltenen Regeln aus und akkumuliert
 * anschließend die abgeleiteten Terme zu Fuzzy-Variablen.
 * <p>
 * Eine Regelbasis ist nicht definiert ({@link #isNull()}), wenn sie keine Regel
 * besitzt.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 * @see Regel
 */
public final class Regelbasis implements Ausdruck<Collection<FuzzyVariable>> {

    public static final Regelbasis NULL = new Regelbasis();

    private final List<Regel> regeln;
    private final BiFunction<Zugehoerigkeit, Zugehoerigkeit, Zugehoerigkeit> akkumulation;

    public Regelbasis(Regel... regeln) {
        this(Arrays.asList(regeln));
    }

    public Regelbasis(List<Regel> regeln) {
        this(Funktionen::maximum, regeln);
    }

    public Regelbasis(BiFunction<Zugehoerigkeit, Zugehoerigkeit, Zugehoerigkeit> akkumulation, Regel... regeln) {
        this(akkumulation, Arrays.asList(regeln));
    }

    public Regelbasis(BiFunction<Zugehoerigkeit, Zugehoerigkeit, Zugehoerigkeit> akkumulation, List<Regel> regeln) {
        Objects.requireNonNull(regeln, "regeln");
        this.regeln = Collections.unmodifiableList(new ArrayList<>(regeln));

        this.akkumulation = Objects.requireNonNull(akkumulation, "akkumulation");
    }

    public boolean isNull() {
        return regeln.isEmpty();
    }

    public boolean nonNull() {
        return !isNull();
    }

    public List<Regel> getRegeln() {
        return regeln;
    }

    @Override
    public Collection<FuzzyVariable> interpretiere(Map<String, ?> kontext) {
        List<TermLiteral> konklussionen = inferenz(kontext);
        return akkumulation(konklussionen);
    }

    private List<TermLiteral> inferenz(Map<String, ?> kontext) {
        return regeln.stream()
                .map(r -> r.interpretiere(kontext))
                .collect(Collectors.toList());
    }

    private Collection<FuzzyVariable> akkumulation(List<TermLiteral> konklusionen) {
        Map<String, List<Term>> fuzzyVariablen = new LinkedHashMap<>();
        konklusionen.forEach(k -> akkumuliere(fuzzyVariablen, k.getFuzzyVariable(), k.getTerm()));
        return fuzzyVariablen.entrySet().stream()
                .map(e -> new FuzzyVariable(e.getKey(), e.getValue()))
                .collect(Collectors.toList());
    }

    private void akkumuliere(Map<String, List<Term>> fuzzyVariablen, String fuzzyVariable, Term term) {
        List<Term> terme = fuzzyVariablen.computeIfAbsent(fuzzyVariable, s -> new ArrayList<>());
        aktualisiereTerme(terme, term);
    }

    private void aktualisiereTerme(List<Term> terme, Term term) {
        int index = -1;
        for (int i = 0; i < terme.size(); i++) {
            if (terme.get(i).getName().equals(term.getName())) {
                index = i;
                break;
            }
        }

        if (index == -1) {
            terme.add(term);
        } else {
            Zugehoerigkeit z1 = terme.get(index).getZugehoerigkeit();
            Zugehoerigkeit z2 = term.getZugehoerigkeit();
            Zugehoerigkeit z = akkumulation.apply(z1, z2);
            term = new Term(term.getName(), z);
            terme.set(index, term);
        }
    }

    @Override
    public String toString() {
        return regeln.stream()
                .map(Regel::toString)
                .reduce((r1, r2) -> r1 + "\n" + r2)
                .orElse("null");
    }

}
