/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import java.util.*;
import java.util.stream.*;

/**
 * Eine Fuzzy-Variable beschreibt die Zugehörigkeiten zu den Fuzzy-Sets einer
 * linguistischen Variable.
 * <p>
 * Ein Fuzzy-Wert ist nicht definiert ({@link #isNull()}), wenn er keinen Term
 * besitzt.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 * @see Term
 */
public final class FuzzyVariable {

    private final String name;
    private final List<Term> terme;

    public FuzzyVariable(String name, Term... terme) {
        this(name, Arrays.asList(terme));
    }

    public FuzzyVariable(String name, List<Term> terme) {
        this.name = Objects.requireNonNull(name, "name");

        Objects.requireNonNull(terme, "terme");
        this.terme = Collections.unmodifiableList(new ArrayList<>(terme));
    }

    public FuzzyVariable mitTerm(Term term) {
        Objects.requireNonNull(term, "term");

        List<Term> terme = new ArrayList<>(this.terme);
        boolean gefunden = false;
        for (int i = 0; i < terme.size(); i++) {
            if (term.getName().equals(terme.get(i).getName())) {
                terme.set(i, term);
                gefunden = true;
            }
        }
        if (!gefunden) {
            terme.add(term);
        }
        return new FuzzyVariable(name, terme);
    }

    public String getName() {
        return name;
    }

    public boolean isNull() {
        return terme.isEmpty();
    }

    public boolean nonNull() {
        return !isNull();
    }

    public List<Term> getTerme() {
        return terme;
    }

    public List<String> getTermnamen() {
        return terme.stream()
                .map(Term::getName)
                .collect(Collectors.toList());
    }

    public Term gibTerm(final String name) {
        Objects.requireNonNull(name, "name");
        return terme.stream()
                .filter(t -> name.equals(t.getName()))
                .findFirst()
                .orElseThrow(() -> new IllegalArgumentException("Die Fuzzy-Variable \"" + this.name + "\" hat keinen Term \"" + name + "\"."));
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        FuzzyVariable that = (FuzzyVariable) o;
        return Objects.equals(name, that.name) &&
                Objects.equals(terme, that.terme);
    }

    @Override
    public int hashCode() {
        return Objects.hash(name, terme);
    }

    @Override
    public String toString() {
        return name + ": " + terme;
    }

}
