/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import java.util.*;
import java.util.function.*;

/**
 * Eine Regel leitet aus vorhandenen Termen einen neuen Term ab.
 * <p>
 * Eine Regel wird in der Form "Wenn Prämisse dann Konklusion mit
 * Sicherheitsgrad" angegeben. Die Prämisse besteht aus belieben Operationen auf
 * Termen die eine Zugehörigkeiten bestimmen. Zusammen mit dem Sicherheitsgrad
 * bildet die bestimmte Zugehörigkeit die Zugehörigkeit für den Term der
 * Konklusion.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public final class Regel implements Ausdruck<TermLiteral> {

    private final Ausdruck<Zugehoerigkeit> praemisse;
    private final TermVariable konklusion;
    private final Sicherheitsgrad sicherheitsgrad;
    private final BiFunction<Zugehoerigkeit, Zugehoerigkeit, Zugehoerigkeit> implikation;

    private Regel(Ausdruck<Zugehoerigkeit> praemisse, TermVariable konklusion, Sicherheitsgrad sicherheitsgrad, BiFunction<Zugehoerigkeit, Zugehoerigkeit, Zugehoerigkeit> implikation) {
        this.praemisse = praemisse;
        this.konklusion = konklusion;
        this.sicherheitsgrad = sicherheitsgrad;
        this.implikation = implikation;
    }

    public static RegelBuilder wenn(Ausdruck<Zugehoerigkeit> praemisse) {
        return new RegelBuilder(praemisse);
    }

    public Ausdruck<Zugehoerigkeit> getPraemisse() {
        return praemisse;
    }

    public TermVariable getKonklusion() {
        return konklusion;
    }

    public Sicherheitsgrad getSicherheitsgrad() {
        return sicherheitsgrad;
    }

    @Override
    public TermLiteral interpretiere(Map<String, ?> kontext) {
        Zugehoerigkeit z;
        try {
            z = implikation.apply(praemisse.interpretiere(kontext), Zugehoerigkeit.von(sicherheitsgrad.getWert()));
        } catch (RuntimeException ignored) {
            z = Zugehoerigkeit.NULL;
        }
        Term term = new Term(konklusion.getTerm(), z);
        return new TermLiteral(konklusion.getFuzzyVariable(), term);
    }

    @Override
    public String toString() {
        return praemisse + " -> " + konklusion + " (" + sicherheitsgrad + ")";
    }

    /**
     * Builder für eine Regel.
     *
     * @author BitCtrl Systems GmbH, Falko Schumann
     */
    public static class RegelBuilder {

        private final Ausdruck<Zugehoerigkeit> praemisse;
        private TermVariable konklusion;
        private Sicherheitsgrad sicherheitsgrad = Sicherheitsgrad.DEFAULT;
        private BiFunction<Zugehoerigkeit, Zugehoerigkeit, Zugehoerigkeit> operator = Funktionen::minimum;

        RegelBuilder(Ausdruck<Zugehoerigkeit> praemisse) {
            this.praemisse = Objects.requireNonNull(praemisse, "praemisse");
        }

        public RegelBuilder dann(TermVariable konklusion) {
            this.konklusion = Objects.requireNonNull(konklusion, "konklusion");
            return this;
        }

        public RegelBuilder mitSicherheitsgrad(Sicherheitsgrad sicherheitsgrad) {
            this.sicherheitsgrad = sicherheitsgrad;
            return this;
        }

        public RegelBuilder benutzeFuerImplikation(BiFunction<Zugehoerigkeit, Zugehoerigkeit, Zugehoerigkeit> operator) {
            this.operator = Objects.requireNonNull(operator, "operator");
            return this;
        }

        public Regel erzeuge() {
            Objects.requireNonNull(konklusion, "konklusion");
            return new Regel(praemisse, konklusion, sicherheitsgrad, operator);
        }

    }

}
